#include <floppy.h>
#include <io.h>
#include <console.h>
#include <timer.h>
#include <memory.h>
#include <idt.h>
#include <pic.h>


const static DmaChannel dmainfo[] = {
   { 0x87, 0x00, 0x01 },
   { 0x83, 0x02, 0x03 },
   { 0x81, 0x04, 0x05 },
   { 0x82, 0x06, 0x07 }
};

/* globals */
static volatile char done = FALSE;
static char dchange = FALSE;
static char motor = FALSE;
static int mtick = 0;
static volatile int tmout = 0;
static char status[7] = { 0 };
static int statsz = 0;
static char sr0 = 0;
static char fdc_track = 0xff;
static DrvGeom geometry = { DG144_HEADS,DG144_TRACKS,DG144_SPT };

static long tbaddr;    /* physical address of track buffer located below 1M */

/* sendbyte() routine from intel manual */
void sendbyte(int byte)
{
   volatile int msr;
   int tmo;
   
   for (tmo = 0;tmo < 128;tmo++) {
      msr = inb(0x3f4);
      if ((msr & 0xc0) == 0x80) {
	 outb(0x3f5,byte);
	 return;
      }
      inb(0x80);   /* delay */
   }
}

/* getbyte() routine from intel manual */
char getbyte()
{
   volatile int msr;
   int tmo;
   
   for (tmo = 0;tmo < 128;tmo++) {
      msr = inb(0x3f4);
      if ((msr & 0xd0) == 0xd0) {
	 return inb(0x3f5);
      }
      inb(0x80);   /* delay */
   }

   return -1;   /* read timeout */
}

void reset(void)
{
   /* stop the motor and disable IRQ/DMA */
   outb(FDC_DOR,0);
   
   mtick = 0;
   motor = FALSE;

   /* program data rate (500K/s) */
   outb(FDC_DRS,0);

   /* re-enable interrupts */
   outb(FDC_DOR,0x0c);

   /* resetting triggered an interrupt - handle it */
   done = TRUE;
   waitfdc(TRUE);

   /* specify drive timings (got these off the BIOS) */
   sendbyte(CMD_SPECIFY);
   sendbyte(0xdf);  /* SRT = 3ms, HUT = 240ms */
   sendbyte(0x02);  /* HLT = 16ms, ND = 0 */
   
   /* clear "disk change" status */
   seek(1);
	 
   recalibrate();

   dchange = FALSE;
}

/* This is the timer (int 1ch) handler */
void time_fd_handler()
{
   if (tmout) --tmout;     /* bump timeout */
   
   if (mtick > 0)
     --mtick;
   else if (!mtick && motor) {
      outb(FDC_DOR,0x0c);  /* turn off floppy motor */
      motor = FALSE;
   }
}

void irq6(void)
{
   /* signal operation finished */
   done = TRUE;
	
   /* EOI the PIC */
   outb(0x20,0x20);
	
	asm("mov %ebp,%esp");
	asm("pop %ebp");
	asm("iret");
}



void fd_init()
{
   int i;

    sendbyte(0x10);

    i = getbyte();

    if (i == 0x80)
	put_string("Floppy: NEC765 controller found [ DONE ]\n");
    else if (i == 0x90)
	put_string("Floppy: 82077 controller found [ DONE ]");
    else
	put_string("Floppy: Enhanced controller found [ DONE ]\n");
		
	tbaddr = 0x00010000	; // WARNING!!! Random Memory.
		
		
	#define LATCH  (PIT_FREQUENCY/INTERVAL)
		
	// Load the IDT timer handler 
	load_idt_entry(38,(unsigned int)&irq6,IDT_CS_SELECTOR,
	IDT_INT_GATE|IDT_SIZE_32|IDT_RING_0|IDT_PRESENT);

	// Enable IRQ of system timer
	enable_irq(6);

	//reset();
	
		
}



void dma_xfer(int channel,long physaddr,int length,char read)
{
   long page,offset;

   /* calculate dma page and offset */
   page = physaddr >> 16;
   offset = physaddr & 0xffff;
   length -= 1;  /* with dma, if you want k bytes, you ask for k - 1 */

   	// Clear interrupts
		asm("cli"); 
   
   /* set the mask bit for the channel */
   outb(0x0a,channel | 4);
   
   /* clear flipflop */
   outb(0x0c,0);
   
   /* set DMA mode (write+single+r/w) */
   outb(0x0b,(read ? 0x48 : 0x44) + channel);
   
   /* set DMA page */
   outb(dmainfo[channel].page,page);
   
   /* set DMA offset */
   outb(dmainfo[channel].offset,offset & 0xff);  /* low byte */
   outb(dmainfo[channel].offset,offset >> 8);    /* high byte */
   
   /* set DMA length */
   outb(dmainfo[channel].length,length & 0xff);  /* low byte */
   outb(dmainfo[channel].length,length >> 8);    /* high byte */
   
   /* clear DMA mask bit */
   outb(0x0a,channel);
   
   // Set interrupts
		asm("sti"); 
}

int waitfdc(char sensei)
{
   tmout = 100;   /* set timeout to 1 second */
     
   /* wait for IRQ6 handler to signal command finished */
   while (!done && tmout)
     ;
  
   /* read in command result bytes */
   statsz = 0;
   while ((statsz < 7) && (inb(FDC_MSR) & (1<<4))) {
      status[statsz++] = getbyte();
   }

   if (sensei) {
      /* send a "sense interrupt status" command */
      sendbyte(CMD_SENSEI);
      sr0 = getbyte();
      fdc_track = getbyte();
   }
   
   done = FALSE;
   
   if (!tmout) {
      /* timed out! */
      if (inb(FDC_DIR) & 0x80)  /* check for diskchange */
				dchange = TRUE;
      
      return FALSE;
   } else
     return TRUE;
}

void block2hts(int block,int *head,int *track,int *sector)
{
   *head = (block % (geometry.spt * geometry.heads)) / (geometry.spt);
   *track = block / (geometry.spt * geometry.heads);
   *sector = block % geometry.spt + 1;
}

/* this turns the motor on */
void motoron(void)
{
   if (!motor) {
      mtick = -1;     /* stop motor kill countdown */
      outb(FDC_DOR,0x1c); 
      sleep(500); /* delay 500ms for motor to spin up */
      motor = TRUE;
   }
}

/* this turns the motor off */
void motoroff(void)
{
   if (motor) {
      mtick = 200;   /* start motor kill countdown: 200 ticks ~ 2s */
   }
}

/* recalibrate the drive */
void recalibrate(void)
{
   /* turn the motor on */
   motoron();
   
   /* send actual command bytes */
   sendbyte(CMD_RECAL);
   sendbyte(0);

   /* wait until seek finished */
   waitfdc(TRUE);
   
   /* turn the motor off */
   motoroff();
}

/* seek to track */
char seek(int track)
{
   if (fdc_track == track)  /* already there? */
     return TRUE;
   
   motoron();
   
   /* send actual command bytes */
   sendbyte(CMD_SEEK);
   sendbyte(0);
   sendbyte(track);

   /* wait until seek finished */
   if (!waitfdc(TRUE))
     return FALSE;     /* timeout! */
 
   /* now let head settle for 15ms */
   sleep(15);
   
   motoroff();
   
   /* check that seek worked */
   if ((sr0 != 0x20) || (fdc_track != track))
		 return FALSE;
   else
     return TRUE;
}

char diskchange(void)
{
   return dchange;
}

/* read block (blockbuff is 512 byte buffer) */
char read_block(int block,char *blockbuff)
{
   return fdc_rw(block,blockbuff,TRUE);
}

/* write block (blockbuff is 512 byte buffer) */
char write_block(int block,char *blockbuff)
{
   return fdc_rw(block,blockbuff,FALSE);
}



char fdc_rw(int block,char *blockbuff,char read)
{
   int head,track,sector,tries; //,i;

   /* convert logical address into physical address */
   block2hts(block,&head,&track,&sector);
   
   /* spin up the disk */
   motoron();

   if (!read && blockbuff) {
      /* copy data from data buffer into track buffer */
      //movedata(_my_ds(),(long)blockbuff,_dos_ds,tbaddr,512);
		 memcpy((void *)tbaddr,blockbuff,512);
   }
   
	 
   for (tries = 0;tries < 3;tries++) {
      /* check for diskchange */
      /*if (!(inb(FDC_DIR) & 0x80)) { // WARNING !!!! Sobra el ! mirar porque
			 dchange = TRUE;
			 seek(1);  //clear "disk change" status 
			 recalibrate();
			 motoroff();				
			 return FALSE;
      }*/
       
      /* move head to right track */
      if (!seek(track)) {
				motoroff();
				return FALSE;
      }
     
      /* program data rate (500K/s) */
      outb(FDC_CCR,0);
      
      /* send command */
      if (read) {
	 			dma_xfer(2,tbaddr,512,FALSE);
	 			sendbyte(CMD_READ);
      } else {
	 			dma_xfer(2,tbaddr,512,TRUE);
	 			sendbyte(CMD_WRITE);
      }
      
      sendbyte(head << 2);
      sendbyte(track);
      sendbyte(head);
      sendbyte(sector);
      sendbyte(2);               /* 512 bytes/sector */
      sendbyte(geometry.spt);
      if (geometry.spt == DG144_SPT)
				sendbyte(DG144_GAP3RW);  /* gap 3 size for 1.44M read/write */
      else
				sendbyte(DG168_GAP3RW);  /* gap 3 size for 1.68M read/write */
      sendbyte(0xff);            /* DTL = unused */
      
      /* wait for command completion */
      /* read/write don't need "sense interrupt status" */
      if (!waitfdc(FALSE))
					return FALSE;   /* timed out! */
      
      if ((status[0] & 0xc0) == 0) break;   /* worked! outta here! */
   
      recalibrate();  /* oops, try again... */
   }
   
   /* stop the motor */
   motoroff();

   if (read && blockbuff) {
      /* copy data from track buffer into data buffer */
      //movedata(_dos_ds,tbaddr,_my_ds(),(long)blockbuff,512);
		 memcpy(blockbuff,(void *)tbaddr,512);
   }

   /*printf("status bytes: ");
   for (i = 0;i < statsz;i++)
     printf("%02x ",status[i]);

   printf("\n");*/

 
   return (tries != 3);
}
